<?php

declare(strict_types=1);

namespace Drupal\ai_agents_test\Entity;

use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\ai_agents_test\AgentTestInterface;
use Drupal\user\EntityOwnerTrait;

/**
 * Defines the agent test entity class.
 *
 * @ContentEntityType(
 *   id = "ai_agents_test",
 *   label = @Translation("Agent Test"),
 *   label_collection = @Translation("Agent Tests"),
 *   label_singular = @Translation("agent test"),
 *   label_plural = @Translation("agent tests"),
 *   label_count = @PluralTranslation(
 *     singular = "@count agent tests",
 *     plural = "@count agent tests",
 *   ),
 *   handlers = {
 *     "list_builder" = "Drupal\ai_agents_test\AgentTestListBuilder",
 *     "views_data" = "Drupal\views\EntityViewsData",
 *     "access" = "Drupal\ai_agents_test\AgentTestAccessControlHandler",
 *     "form" = {
 *       "add" = "Drupal\ai_agents_test\Form\AgentTestForm",
 *       "edit" = "Drupal\ai_agents_test\Form\AgentTestForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm",
 *       "delete-multiple-confirm" = "Drupal\Core\Entity\Form\DeleteMultipleForm",
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider",
 *     },
 *   },
 *   base_table = "ai_agents_test",
 *   admin_permission = "administer ai_agents_test",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "uuid" = "uuid",
 *     "owner" = "uid",
 *   },
 *   links = {
 *     "collection" = "/admin/content/ai-agents-test",
 *     "add-form" = "/admin/content/ai-agents-test/add",
 *     "canonical" = "/admin/content/ai-agents-test/{ai_agents_test}",
 *     "edit-form" = "/admin/content/ai-agents-test/edit/{ai_agents_test}",
 *     "delete-form" = "/admin/content/ai-agents-test/{ai_agents_test}/delete",
 *     "delete-multiple-form" = "/admin/content/ai-agents-test/delete-multiple",
 *   },
 * )
 */
final class AgentTest extends ContentEntityBase implements AgentTestInterface {

  use EntityChangedTrait;
  use EntityOwnerTrait;

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage): void {
    parent::preSave($storage);
    if (!$this->getOwnerId()) {
      // If no owner has been set explicitly, make the anonymous user the owner.
      $this->setOwnerId(0);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type): array {

    $fields = parent::baseFieldDefinitions($entity_type);

    $fields['label'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Label'))
      ->setRequired(TRUE)
      ->setSetting('max_length', 255)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['status'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Status'))
      ->setDefaultValue(TRUE)
      ->setSetting('on_label', 'Enabled')
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'settings' => [
          'display_label' => FALSE,
        ],
        'weight' => 0,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'type' => 'boolean',
        'label' => 'above',
        'weight' => 0,
        'settings' => [
          'format' => 'enabled-disabled',
        ],
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['description'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Description'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('view', [
        'type' => 'text_default',
        'label' => 'above',
        'weight' => 10,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['uid'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Author'))
      ->setSetting('target_type', 'user')
      ->setDefaultValueCallback(self::class . '::getDefaultEntityOwner')
      ->setDisplayOptions('form', [
        'type' => 'entity_reference_autocomplete',
        'settings' => [
          'match_operator' => 'CONTAINS',
          'size' => 60,
          'placeholder' => '',
        ],
        'weight' => 15,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'author',
        'weight' => 15,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Authored on'))
      ->setDescription(t('The time that the agent test was created.'))
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'timestamp',
        'weight' => 20,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'datetime_timestamp',
        'weight' => 20,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the agent test was last edited.'));

    $fields['messages'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Messages'))
      ->setDescription(t('The messages from the agent test as a yaml blob.'))
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'text_textarea',
        'weight' => 35,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'text_default',
        'weight' => 35,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['triggering_instructions'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Triggering Instructions'))
      ->setDescription(t('The instructions that trigger the agent tests. This is only use for getting an overview of what the test does.'))
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'text_textarea',
        'weight' => 5,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'text_default',
        'weight' => 5,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['ai_agent'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('AI Agent'))
      ->setSetting('target_type', 'ai_agent')
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'entity_reference_autocomplete',
        'settings' => [
          'match_operator' => 'CONTAINS',
          'size' => 60,
          'placeholder' => '',
        ],
        'weight' => 25,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => 25,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['rules'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Rules'))
      ->setDescription(t('The rules from the agent test as a yaml blob.'))
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'text_textarea',
        'weight' => 30,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'text_default',
        'weight' => 30,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['config_reset'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Config Reset'))
      ->setDefaultValue(TRUE)
      ->setSetting('on_label', 'Enabled')
      ->setDescription(t('Whether the agent test should reset the config. Only set if needed, since it can be slow. Test group runs might override this.'))
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'settings' => [
          'display_label' => FALSE,
        ],
        'weight' => 35,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'type' => 'boolean',
        'label' => 'above',
        'weight' => 35,
        'settings' => [
          'format' => 'enabled-disabled',
        ],
      ])
      ->setDisplayConfigurable('view', TRUE);

    // The test groups this test is part of.
    $fields['ai_agents_test_group'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Test groups'))
      ->setSetting('target_type', 'ai_agents_test_group')
      ->setCardinality(BaseFieldDefinition::CARDINALITY_UNLIMITED)
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => 40,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => 40,
      ])
      ->setDisplayConfigurable('view', TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function delete() {
    // We also delete any test results associated with this agent test.
    $entities = $this->entityTypeManager()
      ->getStorage('ai_agents_test_result')
      ->loadByProperties(['test_id' => $this->id()]);
    foreach ($entities as $entity) {
      $entity->delete();
    }
    // Delete itself.
    parent::delete();
  }

}
