<?php

declare(strict_types=1);

namespace Drupal\ai_agents_test\Entity;

use Drupal\ai_agents_test\AgentTestInterface;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\ai_agents_test\AiAgentsTestGroupInterface;
use Drupal\user\EntityOwnerTrait;

/**
 * Defines the ai agents test group entity class.
 *
 * @ContentEntityType(
 *   id = "ai_agents_test_group",
 *   label = @Translation("Agent Test Group"),
 *   label_collection = @Translation("Agent Test Groups"),
 *   label_singular = @Translation("agent test group"),
 *   label_plural = @Translation("agent test groups"),
 *   label_count = @PluralTranslation(
 *     singular = "@count agent test groups",
 *     plural = "@count agent test groups",
 *   ),
 *   handlers = {
 *     "list_builder" = "Drupal\ai_agents_test\AiAgentsTestGroupListBuilder",
 *     "views_data" = "Drupal\views\EntityViewsData",
 *     "form" = {
 *       "add" = "Drupal\ai_agents_test\Form\AiAgentsTestGroupForm",
 *       "edit" = "Drupal\ai_agents_test\Form\AiAgentsTestGroupForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm",
 *       "delete-multiple-confirm" = "Drupal\Core\Entity\Form\DeleteMultipleForm",
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider",
 *     },
 *   },
 *   base_table = "ai_agents_test_group",
 *   admin_permission = "administer ai_agents_test_group",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "uuid" = "uuid",
 *     "owner" = "uid",
 *   },
 *   links = {
 *     "collection" = "/admin/content/ai-agents-test/group",
 *     "add-form" = "/admin/content/ai-agents-test/group/add",
 *     "canonical" = "/admin/content/ai-agents-test/group/{ai_agents_test_group}",
 *     "edit-form" = "/admin/content/ai-agents-test/group/{ai_agents_test_group}/edit",
 *     "delete-form" = "/admin/content/ai-agents-test/group/{ai_agents_test_group}/delete",
 *     "delete-multiple-form" = "/admin/content/group/delete-multiple",
 *   },
 * )
 */
final class AiAgentsTestGroup extends ContentEntityBase implements AiAgentsTestGroupInterface {

  use EntityChangedTrait;
  use EntityOwnerTrait;

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage): void {
    parent::preSave($storage);
    if (!$this->getOwnerId()) {
      // If no owner has been set explicitly, make the anonymous user the owner.
      $this->setOwnerId(0);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE) {
    parent::postSave($storage, $update);
    // Make sure to store the group on the test entities.
    foreach ($this->tests as $test_wrapper) {
      $test = $test_wrapper->entity;
      if ($test instanceof AgentTestInterface) {
        // Get the current test group.
        $current_group = $test->get('ai_agents_test_group')->getValue();
        // Add the current group to the test.
        $current_group[] = [
          'target_id' => $this->id(),
        ];
        // Make sure that every target_id is unique.
        $current_group = array_unique(array_column($current_group, 'target_id'));
        // Set the test group on the test.
        $test->set('ai_agents_test_group', $current_group);
        // Save the test.
        $test->save();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type): array {

    $fields = parent::baseFieldDefinitions($entity_type);

    $fields['label'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Label'))
      ->setRequired(TRUE)
      ->setSetting('max_length', 255)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['status'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Status'))
      ->setDefaultValue(TRUE)
      ->setSetting('on_label', 'Enabled')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('view', [
        'type' => 'boolean',
        'label' => 'above',
        'weight' => 0,
        'settings' => [
          'format' => 'enabled-disabled',
        ],
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['description'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Description'))
      ->setDescription(t('A description of the agent test group is suppose to test.'))
      ->setDisplayOptions('form', [
        'type' => 'text_textarea',
        'weight' => 10,
        'rows' => 2,
        'attributes' => [
          'rows' => 2,
        ],
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'type' => 'text_default',
        'label' => 'above',
        'weight' => 10,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['uid'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Author'))
      ->setSetting('target_type', 'user')
      ->setDefaultValueCallback(self::class . '::getDefaultEntityOwner')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'author',
        'weight' => 15,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Authored on'))
      ->setDescription(t('The time that the agent test group was created.'))
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'timestamp',
        'weight' => 20,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the agent test group was last edited.'));

    $fields['tests'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Tests'))
      ->setSetting('target_type', 'ai_agents_test')
      ->setRequired(TRUE)
      ->setCardinality(-1)
      ->setDescription(t('The tests that are part of this group in the order they should run.'))
      ->setCardinality(BaseFieldDefinition::CARDINALITY_UNLIMITED)
      ->setDisplayOptions('form', [
        'type' => 'entity_reference_autocomplete',
        'weight' => 25,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'entity_reference_label',
        'weight' => 25,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['approval_percentage'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Approval Percentage'))
      ->setDescription(t('The percentage of approved tests to total tests to be considered a success.'))
      ->setDefaultValue(90)
      ->setDisplayOptions('form', [
        'type' => 'integer',
        'weight' => 35,
        'settings' => [
          'min' => 0,
          'max' => 100,
        ],
        'attributes' => [
          'min' => 0,
          'max' => 100,
        ],
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'integer',
        'weight' => 35,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['config_reset'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Reset Configuration'))
      ->setDescription(t('This takes a snapshot at the beginning of the test group running and restates it at the end. This can be used when you want to chain dependent tests together.'))
      ->setDefaultValue(FALSE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'weight' => 40,
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'above',
        'type' => 'boolean',
        'weight' => 40,
        'settings' => [
          'format' => 'enabled-disabled',
        ],
      ])
      ->setDisplayConfigurable('view', TRUE);

    return $fields;
  }

}
