<?php

declare(strict_types=1);

namespace Drupal\ai_agents_test\Form;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Yaml\Yaml;

/**
 * Form controller for importing test groups.
 */
final class ImportTestGroupForm extends FormBase {

  /**
   * Constructs a new ImportTestForm object.
   */
  public function __construct(
    protected PrivateTempStoreFactory $tempStore,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected AccountInterface $currentUser,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('tempstore.private'),
      $container->get('entity_type.manager'),
      $container->get('current_user'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'ai_agents_test_import';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    // File upload form with restrictions for YAML files.
    $form['file'] = [
      '#type' => 'file',
      '#title' => $this->t('Upload a test grou file'),
      '#description' => $this->t('Upload a YAML file containing test data. Only .yml or .yaml files are allowed.'),
      '#attributes' => [
        'accept' => '.yml,.yaml',
      ],
    ];
    $form['actions'] = [
      '#type' => 'actions',
    ];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Import'),
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    // Get the temporary file URI.
    /** @var \Symfony\Component\HttpFoundation\FileBag $files */
    $files = $this->getRequest()->files;
    /** @var \Symfony\Component\HttpFoundation\File\UploadedFile $file */
    $file = NULL;
    foreach ($files as $file_objects) {
      $file = $file_objects['file'];
    }
    // If no file.
    if (empty($file)) {
      $form_state->setErrorByName('file', $this->t('Please upload a file.'));
      return;
    }
    // Get the yaml.
    try {
      $yaml = file_get_contents($file->getRealPath());
    }
    catch (\Exception $e) {
      $form_state->setErrorByName('file', $this->t('The file could not be read. Please check the file and try again.'));
      return;
    }
    try {
      // Parse the YAML file.
      $parsed_yaml = Yaml::parse($yaml);
    }
    catch (\Exception $e) {
      $form_state->setErrorByName('file', $this->t('The file is not a valid YAML file. Please check the file and try again.'));
      return;
    }
    // Check so all keys are present.
    $required_keys = ['label', 'description', 'approval_percentage', 'tests', 'config_reset'];
    foreach ($required_keys as $key) {
      if (!array_key_exists($key, $parsed_yaml)) {
        $form_state->setErrorByName('file', $this->t('The file is missing the required key: @key. Please check the file and try again.', ['@key' => $key]));
        return;
      }
    }
    // Also make sure no extra keys are present.
    $allowed_keys = ['label', 'description', 'approval_percentage', 'tests', 'config_reset'];
    foreach ($parsed_yaml as $key => $value) {
      if (!in_array($key, $allowed_keys)) {
        $form_state->setErrorByName('file', $this->t('The file contains an unknown key: @key. Please check the file and try again.', ['@key' => $key]));
        return;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    // Get the temporary file URI.
    /** @var \Symfony\Component\HttpFoundation\FileBag $files */
    $files = $this->getRequest()->files;
    /** @var \Symfony\Component\HttpFoundation\File\UploadedFile $file */
    $file = NULL;
    foreach ($files as $file_objects) {
      $file = $file_objects['file'];
    }
    // Get the yaml.
    $yaml = Yaml::parse(file_get_contents($file->getRealPath()));
    // First create the tests.
    if (empty($yaml['tests']) || !is_array($yaml['tests'])) {
      $form_state->setErrorByName('file', $this->t('The file does not contain any tests. Please check the file and try again.'));
      return;
    }
    $tests = [];
    $testStorage = $this->entityTypeManager->getStorage('ai_agents_test');
    foreach ($yaml['tests'] as $test) {
      // Create the test entity.
      $test_entity = $testStorage->create([
        'label' => $test['label'][0]['value'],
        'status' => $test['status'][0]['value'],
        'description' => $test['description'][0]['value'],
        'uid' => $this->currentUser->id(),
        'created' => $test['created'][0]['value'],
        'changed' => $test['changed'][0]['value'],
        'ai_agent' => $test['ai_agent'][0]['target_id'],
        'messages' => $test['messages'],
        'config_reset' => $test['config_reset'][0]['value'],
        'triggering_instructions' => $test['triggering_instructions'][0]['value'],
        'rules' => $test['rules'],
      ]);
      // Save the test entity.
      $test_entity->save();
      // Add the test entity to the tests array.
      $tests[] = $test_entity;
    }
    // Now create the test group.
    $test_group_storage = $this->entityTypeManager->getStorage('ai_agents_test_group');
    $test_group = $test_group_storage->create([
      'label' => $yaml['label'],
      'description' => $yaml['description'],
      'approval_percentage' => $yaml['approval_percentage'],
      'tests' => $tests,
      'uid' => $this->currentUser->id(),
      'config_reset' => $yaml['config_reset'] ?? FALSE,
    ]);
    // Save the test group entity.
    $test_group->save();
    // Set a message to inform the user.
    $this->messenger()->addMessage($this->t('The test group %label has been imported successfully.', ['%label' => $test_group->label()]));
    // Redirect to the test group view page.
    $form_state->setRedirect('entity.ai_agents_test_group.canonical', ['ai_agents_test_group' => $test_group->id()]);
  }

}
