<?php

namespace Drupal\ai_agents_test\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\ai\AiProviderPluginManager;
use Drupal\ai\OperationType\Chat\ChatInput;
use Drupal\ai\OperationType\Chat\ChatMessage;
use Drupal\ai\OperationType\Chat\Tools\ToolsFunctionInput;
use Drupal\ai\OperationType\Chat\Tools\ToolsInput;
use Drupal\ai\OperationType\Chat\Tools\ToolsPropertyInput;
use Drupal\ai\Service\FunctionCalling\FunctionCallPluginManager;
use Drupal\ai_agents\PluginBase\AiAgentEntityWrapper;
use Drupal\ai_agents\PluginManager\AiAgentManager;
use Drupal\ai_agents_test\AgentTestInterface;
use Drupal\ai_agents_test\AgentTestResultInterface;
use Drupal\ai_agents_test\AgentsParameterTestPluginManager;
use Drupal\ai_agents_test\Traits\RevertConfigTrait;
use Symfony\Component\Yaml\Yaml;

/**
 * Test runner service for AI agents.
 */
class TestRunner {

  use RevertConfigTrait;

  /**
   * The detailed results.
   *
   * @var array
   */
  protected array $detailedResults = [];

  /**
   * Constructs a TestRunner object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\ai_agents\PluginManager\AiAgentManager $aiAgentManager
   *   The AI agent manager service.
   * @param \Drupal\Core\Session\AccountInterface $currentUser
   *   The current user service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The configuration factory service.
   * @param \Drupal\ai\AiProviderPluginManager $aiProvider
   *   The AI provider plugin manager service.
   * @param \Drupal\ai\Service\FunctionCalling\FunctionCallPluginManager $functionCallPluginManager
   *   The function call plugin manager service.
   * @param \Drupal\Core\TempStore\PrivateTempStoreFactory $tempStoreFactory
   *   The private temp store factory service.
   * @param \Drupal\ai_agents_test\AgentsParameterTestPluginManager $testPluginManager
   *   The agents parameter test plugin manager service.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected AiAgentManager $aiAgentManager,
    protected AccountInterface $currentUser,
    protected ConfigFactoryInterface $configFactory,
    protected AiProviderPluginManager $aiProvider,
    protected FunctionCallPluginManager $functionCallPluginManager,
    protected PrivateTempStoreFactory $tempStoreFactory,
    protected AgentsParameterTestPluginManager $testPluginManager,
  ) {
  }

  /**
   * Runs the test for a given test ID.
   *
   * @param int $test_id
   *   The test ID to run.
   * @param string $group_id
   *   If the test should be run in a specific group.
   * @param string $provider_id
   *   The provider ID to use.
   * @param string $model_id
   *   The model ID to use.
   *
   * @return array
   *   The results of the test run.
   */
  public function runTest(int $test_id, string $group_id = '', string $provider_id = "", string $model_id = ""): array {
    // If its a group entity and its the first test we run, we set a checkpoint.
    if (!empty($group_id)) {
      /** @var \Drupal\ai_agents_test\Entity\AgentTestGroupResult $group */
      $group = $this->entityTypeManager->getStorage('ai_agents_test_group_result')->load($group_id);
      // Set checkpoint only if it is the first test in the group.
      if ($group && $group->get('ai_agents_test_results')->isEmpty() && !empty($group->config_reset->value)) {
        $this->setCheckpoint("Backup before the first test in group $group_id.");
        // Store the checkpoint in the temp store.
        $temp_store = $this->tempStoreFactory->get('ai_agents_test');
        $temp_store->set('checkpoint', $this->checkpointStorage);
      }
    }
    // Load the test entity.
    /** @var \Drupal\ai_agents_test\AgentTestInterface $test */
    $test = $this->entityTypeManager->getStorage('ai_agents_test')->load($test_id);
    if (!$test) {
      throw new \InvalidArgumentException(sprintf('Test with ID %d not found.', $test_id));
    }
    // Get the test agent.
    $agent = NULL;
    try {
      /** @var \Drupal\ai_agents\PluginBase\AiAgentEntityWrapper $agent */
      $agent = $this->aiAgentManager->createInstance($test->ai_agent->target_id ?? NULL);
    }
    catch (\Exception $e) {
      throw new \InvalidArgumentException(sprintf('Test with ID %d not found.', $test_id));
    }

    // Create the chat history.
    $messages = Yaml::parse($test->messages->value ?? '');
    if (empty($messages)) {
      throw new \InvalidArgumentException(sprintf('Test with ID %d has no messages.', $test_id));
    }

    // Check so rules are set.
    $rules = Yaml::parse($test->rules->value ?? '');
    if (empty($rules)) {
      throw new \InvalidArgumentException(sprintf('Test with ID %d has no rules.', $test_id));
    }

    $history = [];
    foreach ($messages as $message) {
      $history[] = new ChatMessage($message['role'], $message['content']);
    }
    $chat_input = new ChatInput($history);

    // Set the provider and model if needed.
    if (!empty($provider_id)) {
      $agent->setAiProvider($this->aiProvider->createInstance($provider_id));
    }
    if (!empty($model_id)) {
      $agent->setModelName($model_id);
    }

    // Run the test.
    $agent->setChatInput($chat_input);
    $error_stack = '';
    $text_response = '';
    // Set a checkpoint.
    if ($test->config_reset->value) {
      $this->setCheckpoint("Backup before the $test_id test.");
    }
    try {
      $agent->determineSolvability();
      $text_response = $agent->solve();
      $tools_used = $agent->getToolResults();
    }
    catch (\Throwable $e) {
      // We could not run the test.
      $error_stack = $e->getTraceAsString();
      $error_message = 'Error: ' . $e->getMessage() . ', Line: ' . $e->getLine() . ', File: ' . $e->getFile();
      $tools_used = $agent->getData();
    }

    // Restore the checkpoint.
    if ($test->config_reset->value) {
      $this->revertCheckpoint();
    }

    // If its a PHP error.
    if (isset($error_message)) {
      // Mock a data object.
      $data['entity'] = $this->storeErrorResults($test, $agent, $tools_used, $error_stack, $error_message);
      $data['error'] = $error_message;
      $data['success'] = FALSE;
      $data['results'] = [];
    }
    else {
      $data = $this->checkAgainstRules($rules, $tools_used, $agent, $test, $text_response);
    }

    // If its part of a group, we add the result to the group.
    if (!empty($group_id)) {
      /** @var \Drupal\ai_agents_test\Entity\AgentTestGroupResult $group */
      $group = $this->entityTypeManager->getStorage('ai_agents_test_group_result')->load($group_id);
      if ($group) {
        $current_results = $group->get('ai_agents_test_results')->getValue();
        $current_results[] = [
          'target_id' => $data['entity']->id(),
        ];
        $group->set('ai_agents_test_results', $current_results);
        $group->set('changed', time());
        // If the group results are as many as the tests, we revert.
        if (count($current_results) >= count($group->ai_agents_test_group->entity->tests) && !empty($group->config_reset->value)) {
          // Set the checkpoint from the temp store.
          $temp_store = $this->tempStoreFactory->get('ai_agents_test');
          $checkpoint = $temp_store->get('checkpoint');
          if ($checkpoint) {
            $this->checkpointStorage = $checkpoint;
            // Revert the config.
            $this->revertCheckpoint();
          }
        }
        try {
          $group->save();
        }
        catch (\Exception $e) {
          // We could not save the group.
          return [
            'error' => 'Could not save the group.',
          ];
        }
      }
    }

    return $data;
  }

  /**
   * Test the agent's response against the rules.
   *
   * @param array $rules
   *   The rules to check against.
   * @param array $tools_used
   *   The tools used by the agent.
   * @param \Drupal\ai_agents\PluginBase\AiAgentEntityWrapper $ai_agent
   *   The AI agent entity wrapper.
   * @param \Drupal\ai_agents_test\AgentTestInterface $agent_test
   *   The agent test entity.
   * @param string $text_response
   *   The text response from the agent.
   *
   * @return array
   *   The results of the check.
   */
  public function checkAgainstRules(array $rules, array $tools_used, AiAgentEntityWrapper $ai_agent, AgentTestInterface $agent_test, string $text_response): array {
    $results = [];
    $success = TRUE;
    foreach ($rules as $rule_key => $rule) {
      // Special rules for response checking.
      if ($rule_key == 'response_rule' && $rule) {
        $result = $this->testUsingLlm($rule, $text_response);
        $results['results'][$rule_key] = [
          'result' => $result['result'],
          'label' => 'Response check',
          'message' => $result['reasoning'],
        ];
        $this->detailedResults[] = $this->createDetailedResultRow($result['result'], 'Agent Response', 'LLM check', $rule, $text_response, $result['reasoning']);
        if ($result['result'] === FALSE) {
          $success = FALSE;
        }
      }
      // We might have garbage.
      if (!isset($rule['tool_rules'])) {
        continue;
      }
      // Set success by default.
      $results['results'][$rule_key] = [
        'result' => TRUE,
        'label' => $rule['label'],
        'message' => 'Everything worked as expected.',
      ];
      $tool_name = $this->functionCallPluginManager->getDefinition($rule['tool'])['name'] ?? $rule['tool'];
      // First do the tool rules test.
      if ($rule['tool_rules'] === 'should_be_run') {
        if (!$this->checkToolUsed($rule['tool'], $tools_used)) {
          $success = FALSE;
          $results['results'][$rule_key]['result'] = FALSE;
          $results['results'][$rule_key]['message'] = sprintf('Tool %s was not used.', $tool_name);
          $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Tool should be run', 'Did run', 'Did not run', sprintf('Tool %s was not used.', $rule['tool']));
        }
        else {
          $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Tool should be run', 'Did run', 'Did run', '');
        }
        // Check if the run order was correct.
        $run_order = $this->checkRunOrder($rule, $tools_used);
        if ($run_order['result'] === FALSE) {
          $success = FALSE;
          $results['results'][$rule_key]['result'] = FALSE;
          $results['results'][$rule_key]['message'] = $run_order['message'];
        }
        // Check the parameters combination.
        $parameter_results = $this->checkParameterValue($rule['parameters'], $rule['tool'], $tools_used);
        if ($parameter_results['result'] === FALSE) {
          $success = FALSE;
          $results['results'][$rule_key]['result'] = FALSE;
          // Combine the messages.
          $results['results'][$rule_key]['message'] = implode("\n", $parameter_results['message']);
        }
      }
      elseif ($rule['tool_rules'] === 'should_not_be_run') {
        if ($this->checkToolUsed($rule['tool'], $tools_used)) {
          $success = FALSE;
          $results['results'][$rule_key]['result'] = FALSE;
          $results['results'][$rule_key]['message'] = sprintf('Tool %s was used.', $tool_name);
          $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Tool should not be run', 'Did not run', 'Did run', sprintf('Tool %s was used.', $rule['tool']));
        }
        else {
          $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Tool should not be run', 'Did not run', 'Did not run', '');
        }
      }
    }

    // Save the results.
    $entity = $this->storeTestResults($results, $agent_test, $ai_agent, $success, $tools_used, $text_response);
    $results['entity'] = $entity;
    return $results;
  }

  /**
   * Store the test results.
   *
   * @param array $results
   *   The results of the test.
   * @param \Drupal\ai_agents_test\AgentTestInterface $agent_test
   *   The agent test entity.
   * @param \Drupal\ai_agents\PluginBase\AiAgentEntityWrapper $ai_agent
   *   The AI agent entity wrapper.
   * @param bool $success
   *   Whether the test was successful.
   * @param array $tools_used
   *   The tools used by the agent.
   * @param string $text_response
   *   The text response from the agent.
   *
   * @return \Drupal\ai_agents_test\AgentTestResultInterface|null
   *   The test result entity or NULL if it could not be saved.
   */
  public function storeTestResults(array $results, AgentTestInterface $agent_test, AiAgentEntityWrapper $ai_agent, bool $success, array $tools_used, string $text_response): ?AgentTestResultInterface {
    $storage = $this->entityTypeManager->getStorage('ai_agents_test_result');
    // Load the config of the ai_agent.
    $config = $this->configFactory->get('ai_agents.ai_agent.' . $ai_agent->getId());
    $agent_config = Yaml::dump($config->getRawData(), 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK);
    $test_result = $storage->create([
      'label' => $agent_test->label(),
      'uid' => $this->currentUser->id(),
      'created' => time(),
      'changed' => time(),
      'result' => $success ? 'success' : 'failure',
      'test_id' => $agent_test->id(),
      'summary_results' => Yaml::dump($results['results'], 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'detailed_results' => Yaml::dump($this->detailedResults, 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'system_prompt' => $ai_agent->getAiProvider()->getChatSystemRole(),
      'agent_configuration' => $agent_config,
      'agent_version_id' => md5($agent_config),
      'chat_provider' => $ai_agent->getAiProvider()->getPluginId(),
      'model' => $ai_agent->getModelName(),
      'provider_configuration' => Yaml::dump($ai_agent->getAiConfiguration(), 2, 10, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'tools' => Yaml::dump($this->prepareToolResults($tools_used), 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'response' => $text_response,
    ]);
    try {
      $test_result->save();
    }
    catch (\Exception $e) {
      return NULL;
    }

    return $test_result;
  }

  /**
   * Store the error result, when a PHP error occurs.
   *
   * @param \Drupal\ai_agents_test\AgentTestInterface $agent_test
   *   The agent test entity.
   * @param \Drupal\ai_agents\PluginBase\AiAgentEntityWrapper $ai_agent
   *   The AI agent entity wrapper.
   * @param array $tool_results
   *   The tool results.
   * @param string $error_stack
   *   The error stack.
   * @param string $error_message
   *   The error message.
   */
  public function storeErrorResults(AgentTestInterface $agent_test, AiAgentEntityWrapper $ai_agent, array $tool_results, string $error_stack, string $error_message): ?AgentTestResultInterface {
    $storage = $this->entityTypeManager->getStorage('ai_agents_test_result');
    // Load the config of the ai_agent.
    $config = $this->configFactory->get('ai_agents.ai_agent.' . $ai_agent->getId());
    $agent_config = Yaml::dump($config->getRawData(), 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK);
    $test_result = $storage->create([
      'label' => $agent_test->label(),
      'uid' => $this->currentUser->id(),
      'created' => time(),
      'changed' => time(),
      'result' => 'failure',
      'test_id' => $agent_test->id(),
      'summary_results' => Yaml::dump([], 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'detailed_results' => Yaml::dump([], 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'system_prompt' => $ai_agent->getAiProvider()->getChatSystemRole(),
      'error_stack' => $error_stack,
      'error_message' => $error_message,
      'agent_configuration' => $agent_config,
      'agent_version_id' => md5($agent_config),
      'chat_provider' => $ai_agent->getAiProvider()->getPluginId(),
      'model' => $ai_agent->getModelName(),
      'provider_configuration' => Yaml::dump($ai_agent->getAiConfiguration(), 2, 10, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
      'tools' => Yaml::dump($this->prepareToolResults($tool_results), 10, 2, Yaml::DUMP_MULTI_LINE_LITERAL_BLOCK),
    ]);
    try {
      $test_result->save();
    }
    catch (\Exception $e) {
      return NULL;
    }

    return $test_result;
  }

  /**
   * Create detailed results row.
   *
   * @param bool $result
   *   The result of the test.
   * @param string $object
   *   The object that was tested.
   * @param string $type
   *   The type of test.
   * @param string $expected
   *   The expected result.
   * @param string $actual
   *   The actual result.
   * @param string $details
   *   The details of the test.
   */
  public function createDetailedResultRow(bool $result, string $object, string $type, string $expected, string $actual, string $details): array {
    return [
      'result' => $result,
      'object' => $object,
      'type' => $type,
      'expected' => $expected,
      'actual' => $actual,
      'details' => $details,
    ];
  }

  /**
   * Prepare tool results for storage.
   *
   * @param array $tools_used
   *   The tools used by the agent.
   *
   * @return array
   *   The tools used by the agent, in a format that can be stored in YAML.
   */
  public function prepareToolResults(array $tools_used): array {
    $results = [];
    foreach ($tools_used as $tool) {
      $results[] = [
        'plugin_id' => $tool->getPluginId(),
        'context_values' => $tool->getContextValues(),
        'result' => $tool->getReadableOutput(),
      ];
    }
    return $results;
  }

  /**
   * Check that a tool was used.
   *
   * @param string $tool_id
   *   The tool to check for.
   * @param array $tools_used
   *   The tools used by the agent.
   *
   * @return bool
   *   TRUE if the tool was used, FALSE otherwise.
   */
  public function checkToolUsed(string $tool_id, array $tools_used): bool {
    foreach ($tools_used as $tool) {
      if ($tool->getPluginId() === $tool_id) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Check parameter value.
   *
   * @param array $parameters
   *   The rule to check against.
   * @param string $tool_id
   *   The tool ID to check against.
   * @param array $tools_used
   *   The tools used by the agent.
   *
   * @return array
   *   The results of the check with a failure reason.
   */
  public function checkParameterValue(array $parameters, string $tool_id, array $tools_used): array {
    // Go through each of the tools and see if all parameters worked.
    $errors = [];
    $detailed_results = [];
    /** @var \Drupal\ai\Service\FunctionCalling\ExecutableFunctionCallInterface $tool */
    foreach ($tools_used as $tool_key => $tool) {
      // Set the values for this tool.
      $values = [];
      foreach ($tool->getContextValues() as $key => $value) {
        $values[$key] = $value;
      }
      if ($tool->getPluginId() === $tool_id) {
        $errors[$tool_key] = [];
        $tool_name = $this->functionCallPluginManager->getDefinition($tool_id)['name'] ?? $tool_id;
        // Check if the parameters are correct.
        foreach ($parameters as $parameter_key => $parameter) {
          if ($parameter['type_of_check'] === 'empty') {
            continue;
          }
          $runner = $this->testPluginManager->createInstance($parameter['type_of_check']);
          $runner->runTest($tool, $values, $parameter_key, $parameter, $tool_name);
          if ($runner->getStatus() === FALSE) {
            $errors[$tool_key][] = implode("\n", $runner->getErrors());
            $detailed_results[$tool_key] = array_merge($runner->getDetails(), $detailed_results[$tool_key] ?? []);
          }
          else {
            $detailed_results[$tool_key] = array_merge($runner->getDetails(), $detailed_results[$tool_key] ?? []);
          }
        }
      }
    }

    // If the errors is empty, we are good.
    if (empty($errors)) {
      return [
        'result' => TRUE,
      ];
    }

    // If errors are empty, we are good.
    foreach ($errors as $tool_key => $error) {
      if (empty($error)) {
        // If one is errorless, we say it is successful.
        if (isset($detailed_results[$tool_key])) {
          $this->detailedResults = array_merge($detailed_results[$tool_key], $this->detailedResults);
        }
        return [
          'result' => TRUE,
        ];
      }
    }

    // If we get here, we have errors.
    $key_to_use = NULL;
    $lower_errors = NULL;
    // We take the sub array with the least count of errors.
    foreach ($errors as $tool_key => $error) {
      $amount = count($error);
      if (!$amount) {
        continue;
      }
      if ($lower_errors === NULL || $amount < $lower_errors) {
        $lower_errors = $amount;
        $key_to_use = $tool_key;
      }
    }
    // If we have no errors, we skip.
    if ($key_to_use === NULL) {
      return [
        'result' => TRUE,
      ];
    }

    $this->detailedResults = array_merge($detailed_results[$key_to_use] ?? [], $this->detailedResults ?? []);
    $written_errors = $errors[$key_to_use];

    return [
      'result' => FALSE,
      'message' => $written_errors,
    ];
  }

  /**
   * Checks that the run order of the tools is correct.
   *
   * @param array $rule
   *   The rule to check.
   * @param array $tools_used
   *   The tools used by the agent.
   *
   * @return array
   *   The results of the check with a failure reason.
   */
  public function checkRunOrder(array $rule, array $tools_used): array {
    $tool_name = $this->functionCallPluginManager->getDefinition($rule['tool'])['name'] ?? $rule['tool'];
    // First check what order it is.
    // @todo: Move order to rules as well.
    switch ($rule['tool_order']) {
      case 'any':
        $found_one = FALSE;
        foreach ($tools_used as $order => $tool) {
          $actual_order = $order + 1;
          if ($tool->getPluginId() === $rule['tool'] && !$found_one) {
            $found_one = TRUE;
            $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Run order: any', '*', $actual_order, '');
          }
        }
        return [
          'result' => TRUE,
        ];

      case 'one_of':
        // Get the times it is allowed to be used.
        $allowed = array_map('trim', explode(',', $rule['tool_loop_order']));
        foreach ($tools_used as $order => $tool) {
          $actual_order = $order + 1;
          if ($tool->getPluginId() === $rule['tool']) {
            // Check if the order is allowed.
            if (in_array($actual_order, $allowed)) {
              $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Run order: one of', implode(', ', $allowed), $actual_order, '');
              return [
                'result' => TRUE,
              ];
            }
          }
        }
        // If we get here, it was not allowed.
        $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: one of', implode(', ', $allowed), '', sprintf('Tool %s was not used in the correct order.', $tool_name));
        return [
          'result' => FALSE,
          'message' => sprintf('Tool %s was not used in the correct order.', $rule['tool']),
        ];

      case 'sometime_after_tool':
        // Get the agent it should be after.
        $after_tool = $rule['tool_other_tool_order'];
        $after_tool_name = $this->functionCallPluginManager->getDefinition($after_tool)['name'] ?? $after_tool;
        foreach ($tools_used as $order => $tool) {
          if ($tool->getPluginId() === $after_tool) {
            // If it runs after the tool, we are good.
            $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Run order: sometime after', "Ran after $after_tool_name", "Ran after $after_tool_name", '');
            return [
              'result' => TRUE,
            ];
          }
          elseif ($tool->getPluginId() === $rule['tool']) {
            // If we get here, it was not allowed.
            $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: sometime after', "Ran after $after_tool_name", "Ran before $after_tool_name", sprintf('Tool %s ran before %s.', $tool_name, $after_tool_name));
            return [
              'result' => FALSE,
              'message' => sprintf('Tool %s ran before %s.', $rule['tool'], $after_tool),
            ];
          }
        }
        // If we get here, it was not allowed.
        $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: sometime after', "Ran after $after_tool_name", '', sprintf('Tool %s was not used in the correct order.', $tool_name));
        return [
          'result' => FALSE,
          'message' => sprintf('Tool %s was not used in the correct order.', $rule['tool']),
        ];

      case 'directly_after_tool':
        // Get the agent it should be after.
        $after_tool = $rule['tool_other_tool_order'];
        $after_tool_name = $this->functionCallPluginManager->getDefinition($after_tool)['name'] ?? $after_tool;
        $after_tool_found = FALSE;
        foreach ($tools_used as $order => $tool) {
          if ($tool->getPluginId() === $after_tool) {
            // Check if the order is allowed.
            if ($after_tool_found) {
              $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Run order: directly after', "Ran directly after $after_tool_name", "Ran after $after_tool_name", '');
              return [
                'result' => TRUE,
              ];
            }
            $after_tool_found = TRUE;
          }
          elseif ($tool->getPluginId() === $rule['tool']) {
            // If we get here, it was not allowed.
            $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: directly after', "Ran directly after $after_tool_name", "Ran before $after_tool_name", sprintf('Tool %s ran before %s.', $tool_name, $after_tool_name));
            return [
              'result' => FALSE,
              'message' => sprintf('Tool %s ran before %s.', $rule['tool'], $after_tool),
            ];
          }
        }
        // If we get here, it was not allowed.
        $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: directly after', "Ran directly after $after_tool_name", '', sprintf('Tool %s was not used in the correct order.', $tool_name));
        return [
          'result' => FALSE,
          'message' => sprintf('Tool %s was not used in the correct order.', $rule['tool']),
        ];

      case 'directly_before_tool':
        // Get the agent it should be before.
        $before_tool = $rule['tool_other_tool_order'];
        $before_tool_name = $this->functionCallPluginManager->getDefinition($before_tool)['name'] ?? $before_tool;
        $before_tool_found = FALSE;
        foreach ($tools_used as $order => $tool) {
          if ($tool->getPluginId() === $before_tool) {
            // Check if the order is allowed.
            if ($before_tool_found) {
              $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Run order: directly before', "Ran directly before $before_tool_name", "Ran before $before_tool_name", '');
              return [
                'result' => TRUE,
              ];
            }
            $before_tool_found = TRUE;
          }
          elseif ($tool->getPluginId() === $rule['tool']) {
            // If we get here, it was not allowed.
            $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: directly before', "Ran directly before $before_tool_name", "Ran after $before_tool_name", sprintf('Tool %s ran after %s.', $tool_name, $before_tool_name));
            return [
              'result' => FALSE,
              'message' => sprintf('Tool %s ran after %s.', $rule['tool'], $before_tool),
            ];
          }
        }
        // If we get here, it was not allowed.
        $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: directly before', "Ran directly before $before_tool_name", '', sprintf('Tool %s was not used in the correct order.', $tool_name));
        return [
          'result' => FALSE,
          'message' => sprintf('Tool %s was not used in the correct order.', $rule['tool']),
        ];

      case 'sometime_before_tool':
        // Get the agent it should be before.
        $before_tool = $rule['tool_other_tool_order'];
        $before_tool_name = $this->functionCallPluginManager->getDefinition($before_tool)['name'] ?? $before_tool;
        foreach ($tools_used as $order => $tool) {
          if ($tool->getPluginId() === $rule['tool']) {
            // Check if the order is allowed.
            $this->detailedResults[] = $this->createDetailedResultRow(TRUE, "Tool: $tool_name", 'Run order: sometime before', "Ran before $before_tool_name", "Ran before $before_tool_name", '');
            return [
              'result' => TRUE,
            ];
          }
          elseif ($tool->getPluginId() === $before_tool) {
            // If we get here, it was not allowed.
            $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: sometime before', "Ran before $before_tool_name", "Ran after $before_tool_name", sprintf('Tool %s ran after %s.', $tool_name, $before_tool_name));
            return [
              'result' => FALSE,
              'message' => sprintf('Tool %s ran after %s.', $rule['tool'], $before_tool),
            ];
          }
        }
        // If we get here, it was not allowed.
        $this->detailedResults[] = $this->createDetailedResultRow(FALSE, "Tool: $tool_name", 'Run order: sometime before', "Ran before $before_tool_name", '', sprintf('Tool %s was not used in the correct order.', $tool_name));
        return [
          'result' => FALSE,
          'message' => sprintf('Tool %s was not used in the correct order.', $rule['tool']),
        ];

      default:
        // We should not be here.
        return [
          'result' => TRUE,
        ];

    }
  }

  /**
   * Test using LLM.
   *
   * @param string $test_prompt
   *   The test prompt to use.
   * @param string $value
   *   The value to test against.
   *
   * @return array
   *   The results of the test.
   */
  public function testUsingLlm(string $test_prompt, string $value): array {
    // Get the AI provider.
    $defaults = $this->aiProvider->getSetProvider('chat_with_tools');
    // Create a tool - we do manually, so no tool needs to exist.
    $functions[] = new ToolsFunctionInput('test_results', [
      'description' => 'Test the results.',
      'properties' => [
        new ToolsPropertyInput('result', [
          'description' => 'The result of the test. Should be a string, TRUE or FALSE.',
          'type' => 'string',
          'enum' => [
            'TRUE',
            'FALSE',
          ],
        ]),
        new ToolsPropertyInput('reasoning', [
          'description' => 'The reasoning behind the result.',
          'type' => 'string',
        ]),
      ],
    ]);

    $prompt = "
      You are a test runner for AI agents. You will be given a prompt and a value. The prompt tests the value. You will return the result of the test using the tool.

      The prompt is: $test_prompt

      The value is: $value
    ";
    $input = new ChatInput([
      new ChatMessage('user', $prompt),
    ]);
    $input->setChatTools(new ToolsInput($functions));

    $response = $defaults['provider_id']->chat($input, $defaults['model_id'], [
      'ai_agents_test_runner_result',
    ]);
    $normalized = $response->getNormalized();
    $tools = $normalized->getTools();
    if (!isset($tools[0])) {
      return [
        'result' => FALSE,
        'reasoning' => 'The LLM could not verify anything.',
      ];
    }
    $arguments = $tools[0]->getArguments();
    $return = [
      'result' => FALSE,
      'reasoning' => 'The LLM could not verify anything.',
    ];
    foreach ($arguments as $argument) {
      if ($argument->getName() === 'result') {
        $return['result'] = $argument->getValue() == FALSE || $argument->getValue() == 'FALSE' ? FALSE : TRUE;
      }
      elseif ($argument->getName() === 'reasoning') {
        $return['reasoning'] = $argument->getValue();
      }
    }
    return $return;
  }

}
